import { formatType, forwardArguments, INDENT, INTERNAL, SELF_ARG } from './common';
import { Method } from './method';
import { Struct } from './struct';
import { structuresWithVTableAddress } from './vtable';

// Allow Making Custom Structures
export class CustomWrapperGenerator {
    readonly #struct: Struct;

    // Constructor
    constructor(struct: Struct) {
        this.#struct = struct;
    }

    // Getter
    #getBase() {
        return `${INTERNAL}CustomBase<${this.#struct.name}, ${this.#struct.getVTable().getName()}>`;
    }
    #getVirtualMethods() {
        const out = [];
        const methods = this.#struct.getVTable().getMethods(false);
        for (const method of methods) {
            if (method) {
                out.push(method);
            }
        }
        return out;
    }
    #getName() {
        return 'Custom' + this.#struct.name;
    }

    // Generate Methods
    #generateMethods(methods: Method[], isConstructor: boolean) {
        let out = '';
        for (const method of methods) {
            // Generate Method Signature
            out += INDENT;
            if (isConstructor) {
                out += 'explicit ' + this.#getName();
            } else {
                out += 'virtual ' + formatType(method.returnType) + method.shortName;
            }
            out += method.getArgs(false) + ';\n';
        }
        return out;
    }
    #generateMethodsCode(methods: Method[], isConstructor: boolean) {
        let out = '';
        for (const method of methods) {
            const name = method.shortName;
            // Generate Method Signature
            if (!isConstructor) {
                out += formatType(method.returnType);
            }
            out += this.#getName() + '::';
            out += isConstructor ? this.#getName() : name;
            const args = method.getArgs(false);
            out += args + ' {\n';
            // Generate Method Code
            out += INDENT;
            if (!isConstructor) {
                // Virtual Method
                if (method.doesReturnValue()) {
                    out += 'return ';
                }
                out += `${INTERNAL}VTable::base->${name}`;
            } else {
                // Constructor
                out += method.getDirectCall();
            }
            out += forwardArguments(args, ['this->self']) + ';\n';
            if (isConstructor) {
                // Setup VTable
                out += `${INDENT}${INTERNAL}set_vtable();\n`;
            }
            out += '}\n';
        }
        return out;
    }

    // Generate Header
    generate() {
        let out = '';

        // Check
        if (!structuresWithVTableAddress.includes(this.#struct.name)) {
            throw new Error('Generating A Custom Wrapper Requires The VTable Address To Be Specified');
        }

        // Structure
        out += `struct ${this.#getName()} : ${this.#getBase()} {\n`;

        // Methods
        out += this.#generateMethods(this.#struct.getConstructors(), true);
        out += this.#generateMethods(this.#getVirtualMethods(), false);

        // VTable
        out += 'private:\n';
        out += INDENT + `void ${INTERNAL}patch_vtable(${INTERNAL}VTable *) override final;\n`;
        out += '};\n';

        // Return
        return out;
    }

    // Generate Code
    generateCode() {
        let out = '';

        // Methods
        out += this.#generateMethodsCode(this.#struct.getConstructors(), true);
        const methods = this.#getVirtualMethods();
        out += this.#generateMethodsCode(methods, false);

        // VTable
        const base = this.#getBase();
        out += `void ${this.#getName()}::${INTERNAL}patch_vtable(${base}::${INTERNAL}VTable *vtable) {\n`;
        out += `${INDENT}${base}::${INTERNAL}patch_vtable(vtable);\n`;
        for (const method of methods) {
            const name = method.shortName;
            // Patch Entry
            out += `${INDENT}vtable->${name} = []${method.getArgs()} {\n`;
            out += INDENT + INDENT;
            if (method.doesReturnValue()) {
                out += 'return ';
            }
            out += `custom_get<${this.#getName()}>(${SELF_ARG})->${name}${forwardArguments(method.getArgs(false), [])};\n`;
            out += INDENT + '};\n';
        }
        out += '}\n';

        // Return
        return out;
    }
};